# -*- coding: utf-8 -*-

from django.utils.translation import ugettext as _
from django import forms

from lyra.contrib import drive 
from lyra.contrib.duty import models

def date_range(begin, end):
    assert end >= begin

    cur = begin
    while cur <= end:
        yield cur 
        cur += datetime.timedelta(1)

class WeekFacade(object):
    from django.core import urlresolvers

    def __init__(self, namespace, date):
        self.year, self.week, x_ = date.isocalendar()
        self.namespace = namespace

    def get_absolute_url(self):
        return self.urlresolvers.reverse(
            "%s:browse_week" % self.namespace,
            kwargs={"year": self.year, "week": self.week})
            

class DutyReservation(forms.Form):
    period_start_date = forms.DateField(
        label=_(u"Päivystysjakson ensimmäinen päivämäärä"))
    period_stop_date = forms.DateField(
        label=_(u"Päivystysjakson viimeinen päivämäärä"))
    period_time_start = forms.TimeField(
        label=_(u"Päivystystunnit jaksolla alkaen"))
    period_time_stop = forms.TimeField(
        label=_(u"Päivystystunnit jaksolla kunnes"))

    person = forms.ModelChoiceField(
        queryset=models.DutyPerson.objects.all(),
        label=_(u"Takapäivystäjä"),
        help_text=_(u"Lisää takapäivystäjiä ylläpitopaneelista"))

    def __init__(self, *args, **kwargs):
        self.person = kwargs.pop("person")
        self.namespace = kwargs.pop("namespace")
        self.instance = kwargs.pop("instance")
        self.queryset = kwargs.pop("queryset")
        super(DutyReservation, self).__init__(*args, **kwargs)

    def clean(self):
        cleaned_data = super(DutyReservation, self).clean()

        if all(k in cleaned_data for k in ("period_start_date", 
                                           "period_stop_date",
                                           "period_time_start",
                                           "period_time_stop")):
            if cleaned_data["period_start_date"] > cleaned_data["period_stop_date"]:
                (self._errors
                     .setdefault("period_start_date", self.error_class())
                     .append(_("Varauksen tulee alkaa ennen kuin se loppuu.")))
            if cleaned_data["period_time_start"] > cleaned_data["period_time_stop"]:
                (self._errors
                     .setdefault("period_time_start", self.error_class())
                     .append(_("Varauksen tulee alkaa ennen kuin se loppuu.")))
                

        return cleaned_data        

    def save(self):
        if not self.is_valid:
            raise ValueError
        
        data = self.cleaned_data

        for date in date_range(data["period_start_date"],
                               data["period_stop_date"]):
            models.Reservation.objects.create(
                namespace=self.namespace,
                person=self.person,
                start=datetime.datetime.combine(date, data["period_time_start"]),
                stop=datetime.datetime.combine(date, data["period_time_stop"]),
                description=u"%s puh. tfn. %s" % (data["person"].user.get_full_name(), data["person"].phone))
        
        return WeekFacade(self.namespace, data["period_start_date"])

class DutyApp(drive.DriveApp):
    from django.conf.urls import defaults as urlconf
    
    app_name = "duty"

    def get_app_name(self):
        return _(u"Takapäivystys")

    def user_can_create(self, request):
        return request.user.has_perm("lyra.add_reservation")

    def user_can_edit(self, request, reservation):
        return request.user.has_perm("lyra.change_reservation")

    def user_can_delete(self, request, reservation):
        return request.user.has_perm("lyra.delete_reservation")

    class crud_app(drive.DriveApp.crud_app):
        import datetime

        reservation_form_create = DutyReservation

        def _get_urls(self):
            patterns = super(DutyApp.crud_app, self)._get_urls()
            patterns += self.urlconf.patterns('', *[
                    self.urlconf.url('^%s/(?P<year>\d{4})/(?P<month>\d+)/$' % _("tyoaika"), 
                                     self.month_report, 
                                     name="month_report"),
                    ])
            return patterns

        def month_report(self, request, year, month):
            is_forbidden = self.app.check_forbidden(request, ["view"])
            if is_forbidden:
                return is_forbidden

            year, month = int(year), int(month)
            start = self.datetime.date(year, month, 1)
            if month < 12:
                stop = self.datetime.date(year, month+1, 1)
            else:
                stop = self.datetime.date(year+1, 1, 1)

            reservations = self.app.queryset.date_range(start, stop)
                
            return self.app.get_response(
                request,
                template="month_report",
                context={
                    "reservations": reservations,
                    "date": start
                    })



from lyra import base

from django.utils.translation import ugettext as _
from django.views.generic import edit
from django.views.generic import detail

class CommonContext(object):
    def get_context_data(self, **kwargs):
        data = super(CommonContext, self).get_context_data(**kwargs)
        data.update({"app_name": self.app.get_app_desc()})
        return data

class FormArgumentsMixin(object):
    import datetime

    def get_initial(self):
        initial = super(FormArgumentsMixin, self).get_initial()
        try:
            day = self.request.GET.get("day", u"")
            initial_date = self.datetime.datetime.strptime(
                day, "%Y-%m-%d"
                )
            initial_date = initial_date.replace(hour=8)
            initial.update({
                    "start": initial_date,
                    "stop": initial_date})      
        except ValueError:
            pass

        return initial
            
    def get_form_kwargs(self):
        args = super(FormArgumentsMixin, self).get_form_kwargs()
        args.update({
                "person": self.request.user,
                "namespace": self.app.namespace,
                "queryset": self.app.queryset
                })
        return args


class ReservationCreate(base.AppAwareTemplate, 
                        base.AppAwareSecurityMixin,
                        FormArgumentsMixin,
                        CommonContext,
                        edit.BaseCreateView):
    permissions = ["view", "create"]
    template_name = "reserve"

class ReservationUpdate(base.AppAwareTemplate, 
                        base.AppAwareObjectSecurityMixin, 
                        FormArgumentsMixin,
                        CommonContext,
                        edit.BaseUpdateView):
    permissions = ["view", "edit"]
    template_name = "update"

class ReservationDeletion(base.AppAwareTemplate, 
                          base.AppAwareObjectSecurityMixin,
                          detail.SingleObjectMixin,
                          CommonContext,
                          edit.FormView):
    permissions = ["view", "delete"]
    template_name = "remove"
    
    def get_success_url(self):
        return self.object.get_week_link()

    def get(self, *args, **kwargs):
        self.object = self.get_object()
        return super(ReservationDeletion, self).get(*args, **kwargs)

    def form_valid(self, form):
        self.object = self.get_object()
        self.object.delete()
        return super(ReservationDeletion, self).form_valid(form)

    def form_invalid(self, form):
        self.object = self.get_object()
        return super(ReservationDeletion, self).form_invalid(form)

class ReservationDetail(base.AppAwareTemplate,
                        base.AppAwareObjectSecurityMixin,
                        CommonContext,
                        detail.BaseDetailView):
    template_name = "details"

class Crud(base.SubApp):
    import datetime

    from django import shortcuts
    from django import http
    from django.conf.urls import defaults as urlconf

    from lyra import forms

    reservation_form = forms.ReservationExclusiveEnable
    confirm_form = forms.ConfirmForm

    create = ReservationCreate
    update = ReservationUpdate
    delete = ReservationDeletion
    detail = ReservationDetail

    def _get_urls(self):
        return self.urlconf.patterns('', *[
            self.urlconf.url(
                    '^$', 
                    self.create.as_view(
                        app=self.app,
                        form_class=getattr(self, 
                                           "reservation_form_create", 
                                           self.reservation_form)), 
                    name="reserve"),
            self.urlconf.url(
                    '^(?P<pk>\d+)/$',
                    self.detail.as_view(
                        app=self.app),
                    name="details"),
            self.urlconf.url(
                    '^(?P<pk>\d+)/%s/$' % (_("muokkaa")),
                    self.update.as_view(
                        app=self.app,
                        form_class=getattr(self, 
                                           "reservation_form_edit", 
                                           self.reservation_form)), 
                    name="reserve"),
            self.urlconf.url(
                    '^(?P<pk>\d+)/%s/$' % (_("poista")),
                    self.delete.as_view(
                        app=self.app,
                        form_class=self.confirm_form), 
                    name="remove")])



from lyra import base

from django.utils.translation import ugettext as _

class Browse(base.SubApp):
    import datetime
    import calendar as calendar_mod

    from django.conf.urls import defaults as urlconf
    
    from lyra import dayplanner  
    from lyra import day

    week_display = dayplanner.WeekBrowse
    day_display = dayplanner.DayBrowse
    day_class = day.Day
    
    def __init__(self, *args, **kwargs):
        super(Browse, self).__init__(*args, **kwargs)
        self.calendar = self.calendar_mod.Calendar()

    def _get_urls(self):
        return self.urlconf.patterns('', *[
            self.urlconf.url('^$', self.browse_index, name="browse_index"),
            self.urlconf.url(
                    '^(?P<year>\d{4})/$',
                    self.browse_year, name="browse_year"),
            self.urlconf.url(
                    '^(?P<year>\d{4})/%s(?P<week>\d{1,2})/$' % (_(u"week")),
                    self.week_display.as_view(app=self.app),
                    name="browse_week"),
            self.urlconf.url(
                    '^(?P<year>\d{4})/(?P<month>\d{1,2})/(?P<day>\d{1,2})/$',
                    self.day_display.as_view(app=self.app),
                    name="browse_day"),
            self.urlconf.url(
                    '^(?P<year>\d{4})/(?P<month>\d{1,2})/',
                    self.browse_month, name="browse_month"),])

    def browse_index(self, request):
        is_forbidden = self.app.check_forbidden(request, ["view"])
        if is_forbidden:
            return is_forbidden

        this_year = self.datetime.date.today().year
        next_year_date = self.datetime.datetime(this_year+1, 1, 1)
        all_years = list(self.app.queryset.get_years())
        if next_year_date not in all_years:
            all_years.append(next_year_date)
            all_years.sort()

        return self.app.get_response(
            request,
            template="index", 
            context={
                "years": [
                    {"year": year,
                     "link": self.app.reverse("browse_year", {
                                "year": year.year}),
                     "reservation_count": self.app.queryset.year(year.year).count()}
                    for year in all_years],
                "app_name": self.app.get_app_desc(), 
                },)

    def browse_year(self, request, year):
        is_forbidden = self.app.check_forbidden(request, ["view"])
        if is_forbidden:
            return is_forbidden
        year = int(year)
        
        return self.app.get_response(
            request,
            template="browse_year",
            context={
                "year": year,
                "next_year_link": self.app.reverse("browse_year", {"year": year+1}),
                "prev_year_link": self.app.reverse("browse_year", {"year": year-1}),
                "months": [
                    {"date": self.datetime.date(year, m, 1),
                     "link": self.app.reverse("browse_month", {
                                "month": m, 
                                "year": year}),
                     "reservation_count": self.app.queryset.month(year, m).count()}
                    for m 
                    in range(1, 13)],
                "app_name": self.app.get_app_desc(),  
                "index_link": self.app.reverse("browse_index"),
             },)

    def browse_month(self, request, year, month):
        is_forbidden = self.app.check_forbidden(request, ["view"])
        if is_forbidden:
            return is_forbidden

        def getweek(week):
            date = week[0]
            year, week, weekday = date.isocalendar()
            return week

        def getyear(week):
            date = week[0]
            year, week, weekday = date.isocalendar()
            return year

        year, month = int(year), int(month)

        today = self.datetime.date.today()
        weeks = [{"week": getweek(week),
                  "days": [self.day_class({
                        "date": d, 
                        "events": self.app.queryset.get_for_date(date=d),
                        "is_today": d == today,
                        "is_offmonth": d.month != month})
                           for d in week],
                  "link": self.app.reverse("browse_week", {
                        "year": getyear(week),
                        "week": getweek(week)})} 
                 
                 for week in self.calendar.monthdatescalendar(year, month)]

        next_month = month + 1
        next_year = year
        if next_month > 12:
            next_month = 1
            next_year = year + 1 
        
        prev_month = month - 1
        prev_year = year
        if prev_month < 1:
            prev_month = 12
            prev_year = year - 1 

        return self.app.get_response(
            request,
            template="browse_month",
            context={
                "weeks": weeks,
                "year": year,
                "month": self.datetime.date(year, month, 1),
                "prev_month_link": self.app.reverse("browse_month", {
                        "month": prev_month,
                        "year": prev_year}),
                "prev_year": prev_year,
                "prev_month": prev_month,
                "next_month_link": self.app.reverse("browse_month", {
                        "month": next_month,
                        "year": next_year}),
                "next_year": next_year,
                "next_month": next_month,
                "reserve_link": self.app.reverse("reserve"),
                "year_link": self.app.reverse("browse_year", {"year": year}),
                "app_name": self.app.get_app_desc(),
                },)                



# -*- encoding: utf-8 -*-

import datetime


from django.db import models
from django.db.models import Q
from django.utils.translation import ugettext_lazy as _

from lyra import base

STYLE_CHOICES = (
    ("yellow", _(u"Yellow")),
    ("green", _(u"Green")),
    ("blue", _(u"Blue")),
    ("red", _(u"Red")),
)


class ReservationQuerySet(base.QuerySet):
    def get_for_date(self, date):
        return self.filter(
            Q(start__year=date.year,
                start__month=date.month, 
                start__day=date.day)
            | Q(stop__year=date.year,
                stop__month=date.month,
                stop__day=date.day))
    
    def get_years(self):
        return self.dates("start", "year")

    def get_months(self, year):
        return self.filter(start__year=year).dates("start", "month")
    
    def date_range(self, start_date, stop_date):
        return self.filter(
                # kokonaan sisällä
                Q(start__gte=start_date,
                    stop__lte=stop_date)
                # alkaa ennen ja loppuu alun jälkeen
                | Q(start__lt=start_date,
                    stop__gt=start_date)
                # alkaa ennen loppua ja loppuu lopun jälkeen
                | Q(start__lt=stop_date,
                    stop__gt=stop_date))

    def would_conflict(self, start_date, stop_date):
        return self.date_range(start_date, stop_date).count()

    def month(self, year, month):
        month_start = datetime.date(year, month, 1)
        next_month = month + 1
        if next_month > 12:
            next_month = 1
            next_year = year + 1
        else:
            next_year = year
        month_stop = datetime.date(next_year, next_month, 1) - datetime.timedelta(1)
        return self.date_range(month_start, month_stop)

    def year(self, year):
        year_start = datetime.date(year, 1, 1)
        year_stop = datetime.date(year+1, 1, 1) - datetime.timedelta(1)
        return self.date_range(year_start, year_stop)

class Reservation(models.Model):
    namespace = models.CharField(max_length=64, choices=base.choices())
    person = models.ForeignKey('auth.User')
    person_behalf = models.CharField(
        max_length=255, 
        blank=True,
        verbose_name=_(u"In behalf of"),
        help_text=_(u"Only if reserving by request of someone else"))
    start = models.DateTimeField(verbose_name=_(u"Reservation begins"))
    stop = models.DateTimeField(verbose_name=_(u"Reservation ends"))
    description = models.CharField(
        max_length=140,
        verbose_name=_(u"Short description"),
        help_text=_(u"Shows up on listings"))
    long_description = models.TextField(
        blank=True,
        verbose_name=_(u"Long description"),
        help_text=_(u"Shows up only on detailed view"))
    long_description_markup = models.CharField(
        default="raw",
        max_length=64)
    style = models.CharField(
        verbose_name=_(u"Style"),
        help_text=_(u"Used to distinguish the reservation"),
        max_length=32,
        choices=STYLE_CHOICES,
        default=STYLE_CHOICES[0][0])

    objects = ReservationQuerySet.as_manager()

    @models.permalink
    def get_absolute_url(self):
        return ("%s:details" % self.namespace,
                (),
                {"pk": self.pk})

    @models.permalink
    def get_week_link(self):
        year, week, weekday = self.start.isocalendar()
        return ("%s:browse_week" % self.namespace, 
                (), 
                {"year": year, "week": week})

    def get_creator_name(self):
        if self.person_behalf:
            return _(u"%(behalf)s (tehnyt: %(person)s)") % {
                "behalf": self.person_behalf, 
                "person": self.person.get_full_name()}
        return self.person.get_full_name()

    def one_day(self):
        return self.start.date() == self.stop.date()

    def is_next_week(self):
        today = datetime.date.today()
        date = self.start.date()

        if date < today:
            return False

        monday = today - datetime.timedelta(today.weekday())
        saturday = monday + datetime.timedelta(5)

        return date > saturday
        
    class Meta:
        verbose_name = _(u"Reservation")
        verbose_name_plural = _(u"Reservations")
        ordering = ('start', 'id')

base.clear_choices(Reservation)
register_app = base.make_registerer(Reservation)
ReservationQuerySet.register_app = staticmethod(register_app)
        



import functools
import itertools

from django.db import models
from django.views.generic import base as views_base
from django.template import loader
from django import template as template_module
from django import http

_namespace_registry = {}

class TemplatePrefixMeta(type):
    def __new__(meta, classname, bases, attrs):
        if attrs.get("app_name") and not attrs.get("template_prefix"):
            attrs["template_prefix"] = attrs["app_name"]
        return type.__new__(meta, classname, bases, attrs)
    
class HasUrls(object):   
    @property
    def urls(self):
        patterns = self._get_urls()
        return patterns, self.app_name, self.namespace

    def _get_urls(self):
        raise NotImplementedError
     
class SubApp(HasUrls):
    @property
    def urls(self):
        return self._get_urls()

    def __init__(self, app):
        self.app = app

class App(HasUrls):
    __metaclass__ = TemplatePrefixMeta

    from django.core import urlresolvers

    app_name = None
    extra_context = {}
    app_dict = {}
    template_prefix = None

    def __init__(self, namespace=None, app_name=None,
                 app_namespace=None, template_prefix=None, 
                 parent=None):
        if self.app_name is None and app_name is None:
            raise ValueError("Must define app_name in either "
                             "class scope or in instantiation")
        if app_name is not None:
            self.app_name = app_name 

        if template_prefix:
            self.template_prefix = template_prefix

        self.namespace = namespace or self.app_name
        self.app_namespace = app_namespace or self.namespace

        self.__class__.app_dict.setdefault(self.app_name, []).append(self)
        _namespace_registry[self.namespace] = self
        self.parent = parent

    @staticmethod
    def get_by_namespace(self, namespace):
        return _namespace_registry.get(namespace)

    def get_template_names(self, template_select, 
                           denominator=None, extensions=["html"]):
        some_prefixes = [getattr(p, "template_prefix", None) 
                         for p 
                         in self.__class__.__mro__]
        prefixes = [self.template_prefix] + list(filter(bool, some_prefixes))

        candidates = []

        for (prefix, template, ext
             ) in itertools.product(prefixes, template_select, extensions):
            if denominator is not None:
                candidates.append(
                    "%(app_name)s/custom/%(denominator)s/%(template)s.%(ext)s" % {
                        "app_name": prefix,
                        "denominator": denominator,
                        "template": template,
                        "ext": ext}
                    )
                
            candidates.append(
                "%(app_name)s/%(template)s.%(ext)s" % {
                    "app_name": prefix,
                    "template": template,
                    "ext": ext},
                )

        for (template, ext 
             ) in itertools.product(template_select, extensions):    
            candidates.append(
                "%(template)s.%(ext)s" % {
                    "template": template,
                    "ext": ext,
                    }
                )
            
        return candidates
        

    def get_response(self, request, context={}, template=None,
                     denominator=None, extend_template=None,
                     template_select=None):

        if extend_template is None:
            extend_template = getattr(self, "base_template", "base")

        if template and template_select is None:
            template_select = [template]
        elif template_select is None:
            raise ValueError("provide one of template or "
                             "template_select, not both or neither")
        
        tpl = loader.select_template(
            self.get_template_names(template_select, denominator))

        base_context = dict(
            self.extra_context,
            **{
                "base": loader.select_template(
                    self.get_template_names([extend_template], denominator))}
            )
        context = dict(base_context, **context)
        ctx = template_module.RequestContext(request, context)
        ctx.current_app = self.app_namespace

        return http.HttpResponse(tpl.render(ctx))

    def reverse(self, view_name, kwargs={}):
        return self.urlresolvers.reverse(
            "%s:%s" % (self.namespace, view_name),
            args=(),
            kwargs=kwargs,
            current_app=self.namespace)

# --- Views ---

def requires_app(f):
    @functools.wraps(f)
    def requirer(self, *args, **kwargs):
        if self.app is None:
            raise AttributeError("app aware views need to know"
                                 "the app")
        return f(self, *args, **kwargs)
    return requirer

class AppAwareMixin(object):
    app = None

class AppAwareTemplate(views_base.TemplateResponseMixin):
    app = None

    @requires_app
    def render_to_response(self, *args, **kwargs):
        kwargs["current_app"] = self.app.namespace

        return super(AppAwareTemplate, self).render_to_response(*args, **kwargs)

    def get_context_data(self, **kwargs):
        data = super(AppAwareTemplate, self).get_context_data(**kwargs)
        data.update({
                "base": loader.select_template(self.app.get_template_names(
                    ["base"],
                    denominator=self.get_denominator())),
                })
        return data

    @requires_app
    def get_template_names(self):
        return self.app.get_template_names(
            super(AppAwareTemplate, self).get_template_names(), 
            denominator=self.get_denominator())
    
    def get_denominator(self):
        return None

class AppAwareQuerysetMixin(object):    
    @requires_app
    def get_queryset(self):
        return self.app.queryset

    def get_slug_field():
        return "slug"

    def get_object(self, queryset=None):
        """
        Returns the object the view is displaying.

        By default this requires `self.queryset` and a `pk` or `slug` argument
        in the URLconf, but subclasses can override this to return any object.
        """
        # Use a custom queryset if provided; this is required for subclasses
        # like DateDetailView
        if queryset is None:
            queryset = self.get_queryset()

        # Next, try looking up by primary key.
        pk = self.kwargs.get('pk', None)
        slug = self.kwargs.get('slug', None)
        if pk is not None:
            queryset = queryset.filter(pk=pk)

        # Next, try looking up by slug.
        elif slug is not None:
            slug_field = self.get_slug_field()
            queryset = queryset.filter(**{slug_field: slug})

        # If none of those are defined, it's an error.
        else:
            raise AttributeError(u"Generic detail view %s must be called with "
                                 u"either an object pk or a slug."
                                 % self.__class__.__name__)

        try:
            obj = queryset.get()
        except ObjectDoesNotExist:
            raise Http404(_(u"No %(verbose_name)s found matching the query") %
                          {'verbose_name': queryset.model._meta.verbose_name})
        return obj


class AppAwareSecurityMixin(object):
    permissions = ["view"]

    @requires_app
    def dispatch(self, request, *args, **kwargs):
        is_forbidden = self.app.check_forbidden(
            request, 
            self.permissions)
        if is_forbidden:
            return is_forbidden

        return super(AppAwareSecurityMixin, self).dispatch(request, *args, **kwargs)

class AppAwareObjectSecurityMixin(AppAwareQuerysetMixin):
    permissions = ["view"]

    @requires_app
    def dispatch(self, request, *args, **kwargs):
        self.kwargs = kwargs
        obj = self.get_object()

        is_forbidden = self.app.check_forbidden(
            request, 
            self.permissions,
            obj)
        if is_forbidden:
            return is_forbidden

        return super(AppAwareObjectSecurityMixin, self).dispatch(
            request, *args, **kwargs)

# --- Models ---

class QuerySetManager(models.Manager):
    use_for_related_fields = True
    def __init__(self, qs_class=models.query.QuerySet):
        self.queryset_class = qs_class
        super(QuerySetManager, self).__init__()

    def get_query_set(self):
        return self.queryset_class(self.model)

    def __getattr__(self, attr, *args):
        try:
            return getattr(self.__class__, attr, *args)
        except AttributeError:
            return getattr(self.get_query_set(), attr, *args)

class QuerySet(models.query.QuerySet):
    """Base QuerySet class for adding custom methods that are made
    available on both the manager and subsequent cloned QuerySets"""

    @classmethod
    def as_manager(cls, ManagerClass=QuerySetManager):
        return ManagerClass(cls)

choices_placeholder = ("1", "1")
choices = lambda: [choices_placeholder]

def list_get_name(L, key):
    for i in L:
        if i.name == key:
            return i
    raise IndexError

def clear_choices(model, field_name="namespace"):
    choices = list_get_name(model._meta.fields, field_name).choices
    try:
        del choices[choices.index(choices_placeholder)]
    except ValueError:
        pass

ALL_NAMESPACES = {}

def make_registerer(model, field_name="namespace"):
    def register_app(section_tuple, app_name):
        L = list_get_name(model._meta.fields, field_name).choices
        (section_namespace, section_desc) = section_tuple
        if section_namespace not in [i for i,d in L]:
            L.append(section_tuple)
            ALL_NAMESPACES.setdefault(app_name, {})[section_namespace] = section_desc
    clear_choices(model, field_name)

    return register_app



from django.views.generic import base as views_base

def quarts(time):
    """quarters from 0:00, rounded up
    
    values range from 0 (0:00-0:15) to 95 (23:45-0:00)"""
    return time.hour * 4 + time.minute // 15

def get_weekdays(year, week):
    import datetime
    first_day = datetime.date(year, 1, 1) 
    monday = first_day + datetime.timedelta(
        days=-first_day.weekday(), 
        weeks=week)
    return [monday+datetime.timedelta(days=i) for i in range(7)]

class DayPlanner(views_base.View):
    import datetime

    from lyra import day

    app = None

    QUARTER_HEIGHT = 15 #px
    DISPLAY_WEEKENDS = True
    
    day_class = day.Day

    template_name = "browse_week"

    def render(self, weekdays, extra_context=None, template_name=None):
        if extra_context is None:
            extra_context = {}

        is_forbidden = self.app.check_forbidden(self.request, ["view"])
        if is_forbidden:
            return is_forbidden

        (self.months_mentioned, 
         self.years_mentioned) = self.collect_months_and_years(weekdays)
        
        today = self.datetime.date.today()
        days = [self.day_class({
                    "date": d, 
                    "events": self.process_day_events(
                        d,
                        self.app.queryset.get_for_date(date=d)),
                    "link": self.app.reverse("browse_day", {
                            "year": d.year,
                            "month": d.month,
                            "day": d.day}),
                    "is_today": d == today})
                for d in weekdays]
        
        self.min_quart, self.max_quart = self.get_quart_bounds(days)
        self.business_hours = self.get_business_hours()
            
        if self.business_hours:
            self.days_columns = self.generate_columns(days)
        else:
            self.days_columns = []

        return self.app.get_response(
            self.request,
            template=template_name or self.template_name,
            context=dict({
                "days": self.days_columns,
                "months_mentioned": self.months_mentioned,
                "years_mentioned": self.years_mentioned,
                "business_hours": self.business_hours,
                "hour_height": self.QUARTER_HEIGHT * 4,
                "table_max_height": (self.QUARTER_HEIGHT * 4 
                                     * (len(self.business_hours))),
                "reserve_link": (self.app.user_can_create(self.request) 
                                 and self.app.reverse("reserve")),
                "app_name": self.app.get_app_desc(),}, 
                         **extra_context))

    def annotate_vacancies(self, column):
        new_column = []
        current_quart = self.min_quart

        for event in column:
            event_quarts = quarts(event["stop"]) - quarts(event["start"])
            event.update({
                    "is_vacant": False,
                    "height": event_quarts * self.QUARTER_HEIGHT,
                    "top": ((quarts(event["start"]) - self.min_quart) 
                            * self.QUARTER_HEIGHT),
                    "quarts": event_quarts,
                    })
            new_column.append(event)

        return new_column
    
    def collect_months_and_years(self, weekdays):
        months_mentioned = []
        years_mentioned = []
        for d in (weekdays[0], weekdays[-1]):
            if d.month not in [m["date"].month for m in months_mentioned]:
                months_mentioned.append(
                    {"date": d,
                     "link": self.app.reverse("browse_month", {
                                "year": d.year, 
                                "month": d.month})})
            if d.year not in [y["date"].year for y in years_mentioned]:
                years_mentioned.append(
                    {"date": d,
                     "link": self.app.reverse("browse_year", {
                                "year": d.year})})
                    
        return months_mentioned, years_mentioned

    def get_quart_bounds(self, days):
        min_quart = 24*4
        max_quart = 0

        for day in days:
            min_quart = min([min_quart] + [
                    quarts(e["start"]) for e in day["events"]])
            max_quart = max([max_quart] + [
                    quarts(e["stop"]) for e in day["events"]])
    
        min_quart -= min_quart % 4
        max_quart -= max_quart % 4 - 8
            
        return min_quart, max_quart

    def get_business_hours(self):
        if self.min_quart > self.max_quart:
            # no (well-formed) events
            business_hours = []
        else:
            business_hours = [a % 24 for a in range(self.min_quart//4, 
                                                    self.max_quart//4)]
        return business_hours

    def conflicts(self, start1, start2, end1, end2):
        if start1 < start2 < end1:
            return True
        if start1 < end2 < end1:
            return True
        return False

    def fits(self, list_, event):
        return not any(self.conflicts(event["start"], event["stop"], 
                                      e["start"], e["stop"])
                       for e in list_)

    def split_columns(self, events):
        # list of single empty list to ensure every day 
        # has at least one column
        columns = [[]]
        for event in events:
            for column in columns:
                if self.fits(column, event):
                    column.append(event)
                    break
            else:
                columns.append([event])
                
        return columns        

    def process_day_events(self, date, queryset):
        return [dict(vars(e), **{
                    # multi day events show up multiple times, for each day
                    "start": max([
                            e.start, 
                            self.datetime.datetime.combine(
                                date, 
                                self.datetime.time(hour=0, minute=0, second=0))]),
                    "stop": min([
                            e.stop,
                            self.datetime.datetime.combine(
                                date, 
                                self.datetime.time(hour=23, minute=59, second=59))]),
                    "creator_name": e.get_creator_name(),
                    "update_link": self.app.reverse(
                        "reserve", 
                        kwargs={"pk": e.pk}),
                    "remove_link": self.app.reverse(
                        "remove", 
                        kwargs={"pk": e.pk}),
                    "can_update": self.app.user_can_edit(self.request, e),
                    "can_remove": self.app.user_can_delete(self.request, e),
                    "link": e.get_absolute_url(),
                    "one_day": e.one_day(),
                    "instance": e,
                })
                for e 
                in queryset]

    def generate_columns(self, days):
        for day in days:
            day["event_columns"] = [
                self.annotate_vacancies(column) 
                for column in self.split_columns(day["events"])]
        return days

class WeekBrowse(DayPlanner):
    def get(self, request, year, week, template_name=None, extra_context={}):
        year, week = int(year), int(week)
        weekdays = get_weekdays(year, week)

        if not self.DISPLAY_WEEKENDS:
            weekdays = weekdays[:5]
            
        monday = weekdays[0]
        (prev_year, 
         prev_week, 
         x) = (monday - self.datetime.timedelta(7)).isocalendar()
        (next_year,
         next_week,
         x) = (monday + self.datetime.timedelta(7)).isocalendar()

        return self.render(
            weekdays, 
            extra_context=dict({
                "week": week,
                "year": year,
                "prev_week_link": self.app.reverse("browse_week", {
                        "week": prev_week,
                 "year": prev_year,}),
                "prev_week": prev_week,
                "next_week_link": self.app.reverse("browse_week", {
                        "week": next_week,
                        "year": next_year}),
                "next_week": next_week,
                }, **extra_context),
            template_name=template_name)

class DayBrowse(DayPlanner):
    template_name = "browse_day"

    def get(self, request, year, month, day):
        date = self.datetime.date(int(year), int(month), int(day))

        tomorrow = date + self.datetime.timedelta(1)
        yesterday = date - self.datetime.timedelta(1)
 
        return self.render([date], extra_context={
                "date": date,
                "next_day": tomorrow,
                "next_day_link": self.app.reverse(
                    "browse_day", {"year": tomorrow.year, 
                                   "month": tomorrow.month,
                                   "day": tomorrow.day}),
                "prev_day": yesterday,
                "prev_day_link": self.app.reverse(
                    "browse_day", {"year": yesterday.year,
                                   "month": yesterday.month,
                                   "day": yesterday.day}),
                })




class Day(dict):
    import datetime

    def css_class(self):
        classes = []
        other_days = []
        if self.get("is_offmonth"):
            other_days.append("offmonth")
        if self.is_weekend():
            other_days.append("weekend")
        if self.get("is_today"):
            return "today"
        if other_days:
            classes.append("-".join(other_days))
        if self["date"].toordinal() < self.datetime.date.today().toordinal():
            classes.append("past-day")

        return " ".join(classes)
    
    def is_weekend(self):
        return self["date"].weekday() in [5,6]



# -*- encoding: utf-8 -*-

import calendar
import datetime

from django.utils import translation
from django.utils.translation import ugettext as _
from django.contrib.auth import decorators as auth_decorators
from django.utils.decorators import method_decorator
from django.views.generic import base as views_base

from lyra import base

login_required_m = method_decorator(auth_decorators.login_required)

class PermissionError(Exception):
    def __init__(self, *args, **kwargs):
        self.reason = kwargs.pop("reason", "admin")
        super(PermissionError, self).__init__(*args, **kwargs)

class Lyra(base.App):
    import datetime

    from django import http
    from django import shortcuts
    from django import template
    from django.conf import settings
    from django.conf.urls import defaults as urlconf
    from django.core import urlresolvers
    from django.db.models import Q

    from lyra import models
    from lyra import forms
    from lyra import browse as browse_mod
    from lyra import crud as crud_mod

    model = models.Reservation
    app_name="lyra"

    # subapplications
    browse_class = browse_mod.Browse
    crud_class = crud_mod.Crud

    # errors
    perm_error = PermissionError

    def get_app_desc(self):
        return _(u"Calendar")

    def __init__(self, *args, **kwargs):
        super(Lyra, self).__init__(*args, **kwargs)
        self.model.objects.register_app(
            (self.namespace, self.get_app_desc()), 
            self.app_name)

        self._queryset = None

        self.browse = self.browse_class(app=self)
        self.crud = self.crud_class(app=self)

    @property
    def queryset(self):
        if self._queryset:
            return self._queryset

        qs = self._queryset = self.model.objects.filter(
            namespace=self.namespace)
        return qs
    
    def _get_urls(self):
        patterns = self.urlconf.patterns('', *[
                self.urlconf.url('^$', self.landing, name="landing"),
                self.urlconf.url('^%s/' % _(u"date"), 
                    self.urlconf.include(self.browse.urls)),
                self.urlconf.url(
                    '^%s/' % _(u"reservation"),
                    self.urlconf.include(self.crud.urls))
                ])

        return patterns

    def landing(self, request):
        today = self.datetime.date.today()
        year, week, weekday = today.isocalendar()

        return self.browse.week_display.as_view(app=self)(
            request, 
            year, 
            week, 
            template_name="landing")

    def user_can_create(self, request):
        return request.user.is_authenticated()

    def user_can_edit(self, request, res):
        user = request.user
        return (user.is_authenticated() 
                and user == res.person)
    
    def user_can_delete(self, request, res):
        user = request.user
        return (user.is_authenticated() 
                and user == res.person)

    def user_can_view(self, request, obj=None): 
        return True
    
    def check_forbidden(self, request, perm_list, obj=None):
        def try_meth(m):
            try:
                return m(request, obj)
            except TypeError:
                try:
                    return m(request)
                except TypeError:
                    pass
                raise
        
        for perm in perm_list:
            perm_method = getattr(self, "user_can_%s" % perm)
            if not perm_method:
                raise ValueError("unknown permission %s" % perm)

            try:
                allowed = try_meth(perm_method)
                if not allowed:
                    return self.forbidden(request)
            except self.perm_error, exc:
                return self.forbidden(request, exc.reason)
        
    def forbidden(self, request, reason="admin"):
        return self.get_response(request, 
                                 template="forbidden", 
                                 context={"reason": reason}) 

    def get_a_couple_events(self):
        now = self.datetime.datetime.now()
        return (self.queryset
                .filter((self.Q(start__lte=now)
                         & self.Q(stop__gte=now))
                        | self.Q(start__gt=now))
                .order_by("start"))[:4]



# -*- coding: utf-8 -*-

from django.utils.translation import ugettext as _
from django import forms

from lyra import views
from lyra import dayplanner

class FridayForm(forms.Form):
    lunch_main = forms.CharField(
        label=_(u"Lunch main course"),
        required=False)
    lunch_other = forms.CharField(
        widget=forms.Textarea, 
        required=False,
        label=_(u"Lunch side courses"))

    def __init__(self, *args, **kwargs):
        self.date = kwargs.pop("date")
        self.queryset = kwargs.pop("queryset")
        self.new_object = kwargs.pop("new_object")
        
        kwargs["initial"] = self.get_initial()

        super(FridayForm, self).__init__(*args, **kwargs)

    def get_initial(self):
        lunch = self.get_lunch_reservation()
        return {
            "lunch_main": lunch.description,
            "lunch_other": lunch.long_description}

    def get_lunch_reservation(self):
        start = datetime.datetime.combine(self.date, datetime.time(11, 15))

        try:
            return self.queryset.get(start=start)
        except django_exceptions.ObjectDoesNotExist:
            return self.new_object(start=start, stop=start.replace(hour=12, minute=0))

    def save(self, commit=True):
        res = self.get_lunch_reservation()
        res.description = self.cleaned_data["lunch_main"]
        res.long_description = self.cleaned_data["lunch_other"]
        if commit:
            res.save()
        return [res]

class DayForm(FridayForm):
    dinner_main = forms.CharField(
        label=_(u"Dinner main course"),
        required=False)
    dinner_other = forms.CharField(
        widget=forms.Textarea, 
        required=False,
        label=_(u"Dinner side courses"))

    def get_initial(self):
        dinner = self.get_dinner_reservation()

        return dict(super(DayForm, self).get_initial(),
                    dinner_main=dinner.description,
                    dinner_other=dinner.long_description)

    def get_lunch_reservation(self):
        start = datetime.datetime.combine(self.date, datetime.time(11, 0))

        try:
            return self.queryset.get(start=start)
        except django_exceptions.ObjectDoesNotExist:
            return self.new_object(start=start, stop=start.replace(hour=12, minute=45))

    def get_dinner_reservation(self):
        start = datetime.datetime.combine(self.date, datetime.time(16, 0))

        try:
            return self.queryset.get(start=start)
        except django_exceptions.ObjectDoesNotExist:
            return self.new_object(start=start, stop=start.replace(hour=17))

    def save(self, commit=True):
        reservations = super(DayForm, self).save(commit=False)
        res = self.get_dinner_reservation()
        res.description = self.cleaned_data["dinner_main"]
        res.long_description = self.cleaned_data["dinner_other"]
        reservations.append(res)
        if commit:
            for res in reservations:
                res.save()
        return reservations

class Menu(views.Lyra):
    app_name = "food"
    
    def get_app_name(self):
        return _(u"Menu")

    class crud_app(views.Lyra.crud_app):
        from django import http

        from lyra import models

        day_form = DayForm
        friday_form = FridayForm
        reservation_model = models.Reservation

        def _get_urls(self):
            patterns = super(Menu.crud_app, self)._get_urls()
            patterns += self.urlconf.patterns('', *[
                    self.urlconf.url('^%s/(?P<year>\d{4})/%s(?P<week>\d+)/$' % (_("viikon-varaus"), _("viikko")), 
                                     self.reserve_week,
                                     name="reserve_week"),
                    ])
            return patterns

        
        def reserve_week(self, request, year, week):
            is_forbidden = self.app.check_forbidden(
                request, ["view", "create"])
            if is_forbidden:
                return is_forbidden

            year, week = int(year), int(week)

            weekdays = dayplanner.get_weekdays(year, week)
            monthu = weekdays[:4]
            fri = weekdays[4]

            def new_object(**kwargs):
                our_kwargs = dict(kwargs, **{"person": request.user, 
                                             "namespace": self.app.namespace})
                return self.reservation_model(**our_kwargs)

            if request.method == "POST":
                forms = []
                for i, date in enumerate(monthu):
                    forms.append(self.day_form(request.POST, 
                                               date=date, 
                                               prefix=i, 
                                               queryset=self.app.queryset,
                                               new_object=new_object))
                forms.append(self.friday_form(request.POST, 
                                              date=fri, 
                                              prefix=i+1,
                                              queryset=self.app.queryset,
                                              new_object=new_object))

                if all(f.is_valid() for f in forms):
                    for form in forms:
                        form.save()
                    return self.http.HttpResponseRedirect(
                        self.app.reverse("browse_week", {
                                "year": year, 
                                "week": week}))
            else:
                forms = []
                for i, date in enumerate(monthu):
                    forms.append(self.day_form(
                            date=date, 
                            prefix=i, 
                            queryset=self.app.queryset, 
                            new_object=new_object))
                forms.append(self.friday_form(
                        date=fri,
                        prefix=i+1, 
                        queryset=self.app.queryset, 
                        new_object=new_object))

            return self.app.get_response(
                request,
                template="reserve_week",
                context={"forms": forms,
                         "app_name": self.app.get_app_name()})


    class browser_app(views.Lyra.browse_app):
        class week_display(views.Lyra.browse_app.week_display):
            import datetime

            DISPLAY_WEEKENDS = False
        
            BREAKFAST_EVENTS = {
                0: {"description": u"Kaurapuuro"},
                1: {"description": u"Ruishiutalepuuro",},
                2: {"description": u"Neljänviljanpuuro"},
                3: {"description": u"Ohrahiutalepuuro"},
                4: {"description": u"Mannapuuro"},
                }

            BREAKFAST_COMMON = {
                "long_description": u"Kahvi, tee",
                "start": datetime.time(7),
                "stop": datetime.time(8, 30),
                "style": "yellow",
                }
            
            def process_day_events(self, date, queryset):
                events = (super(FoodSchedule.browser_app.week_display, 
                                self)
                          .process_day_events(date, queryset))

                meaningful_events = [e for e in events 
                                     if e.get("description")
                                     or e.get("long_description")]
                
                if meaningful_events:
                    breakfast = dict(self.BREAKFAST_EVENTS[date.weekday()],
                                     **self.BREAKFAST_COMMON)
                
                    for k in ["start", "stop"]:
                        breakfast[k] = self.datetime.datetime.combine(date, breakfast[k])
                        
                    return [breakfast] + meaningful_events
                return []

            ROW_LABEL_NAMES = ["start", "stop", "label"]
            ROW_LABELS = [
                (datetime.time(7, 0), datetime.time(8, 45), u"Aamiainen"),
                (datetime.time(11, 0), datetime.time(12, 45), u"Lounas"),
                (datetime.time(16, 0), datetime.time(17, 0), u"Päivällinen")
                ]

            def week_table_rows(self, year, week):
                columns = [[None] + [
                    dict(zip(self.ROW_LABEL_NAMES, labels)) 
                    for labels in self.ROW_LABELS]]
                weekdays = dayplanner.get_weekdays(int(year), int(week))[:5]

                for weekday in weekdays:
                    col = [{"date": weekday}]
                    events = self.process_day_events(
                        weekday, 
                        self.app.queryset.get_for_date(weekday))

                    for start, stop, label in self.ROW_LABELS: 
                        es = [e for e in events 
                             if e["start"].time() >= start
                             and e["stop"].time() <= stop]
                        if es:
                            col.append(es[0])
                        else:
                            col.append(None)
                            
                    if all(not cell for cell in col[1:]):
                        col[2] = {
                            "description": _(u"(Päivä jätetty tyhjäksi)")
                            }

                    columns.append(col)

                return zip(*columns)

            def get(self, request, year, week, *args, **kwargs):
                if "print" in request.GET:
                    kwargs["template_name"] = "week_print"
                    kwargs["extra_context"] = {
                        "table_rows": self.week_table_rows(year, week)
                        }

                return super(Lyra.browse_app.week_display, self).get(
                    request, year, week, *args, **kwargs)



from django.utils.translation import ugettext_lazy as _
from django.db import models
    
class DutyPerson(models.Model):
    user = models.ForeignKey('auth.User', 
                             verbose_name=_(u"Käyttäjätunnus"))
    phone = models.CharField(max_length=127, 
                             verbose_name=_(u"Puhelinnumero"))
    accounting_no = models.CharField(
        max_length=127,
        verbose_name=_(u"Tiliöintinumero"),
        default=u"")

    def __unicode__(self):
        return unicode(self.user)

    class Meta:
        verbose_name = _(u"Takapäivystäjä")
        verbose_name_plural = _(u"Takapäivystäjiä")



# -*- encoding: utf-8 -*-

import datetime

from django import forms
from django.db.models import Q
from django.utils.translation import ugettext_lazy as _
from django.core import exceptions as django_exceptions

from lyra import models

class Reservation(forms.ModelForm):
    def __init__(self, *args, **kwargs):
        self.person = kwargs.pop("person")
        self.namespace = kwargs.pop("namespace")
        self.queryset = kwargs.pop("queryset")
        super(Reservation, self).__init__(*args, **kwargs)

    def clean(self):
        cleaned_data = super(Reservation, self).clean()

        if all(k in cleaned_data for k in ("start", "stop")):
            start = cleaned_data["start"]
            stop = cleaned_data["stop"]
            if start > stop:
                (self._errors
                     .setdefault("start", self.error_class())
                     .append(_("The reservation should begin before it ends")))

        return cleaned_data

    def save(self, commit=True, **kwargs):
        obj = super(Reservation, self).save(commit=False, **kwargs)
        
        if not obj.pk:
            obj.person = self.person
            obj.namespace = self.namespace

        if commit:
            obj.save()
        return obj            

    class Meta:
        model = models.Reservation
        exclude = ("namespace", "person", "long_description_markup")
        widgets = {
            "style": forms.Select(attrs={"class": "schedule_style"}),
            }

    class Media:
        js = ("shared/js/sivari.stylepreview.js",)
    
class ReservationExclusive(Reservation):
    def toggle_enabled(self, cleaned_data):
        return (not hasattr(self, "exclusive")
                and cleaned_data.get("exclusive"))

    def clean(self):
        cleaned_data = super(ReservationExclusive, self).clean()

        start_date = cleaned_data.get("start")
        stop_date = cleaned_data.get("stop")
        if start_date and stop_date and self.toggle_enabled(cleaned_data):
            would_conflict = self.queryset.date_range(
                start_date, stop_date)
            if self.instance.pk: 
                would_conflict = would_conflict.exclude(pk=self.instance.pk)
            if would_conflict.count():
                (self._errors
                     .setdefault("start", self.error_class())
                     .append(_(u"The reservation would conflict with %(conflict_count)s "
                               u"other reservations.") % {
                            "conflict_count": would_conflict}))

        return cleaned_data        

class ReservationExclusiveEnable(ReservationExclusive):
    exclusive = forms.BooleanField(
        label=_(u"No overlap"),
        required=False)

class ReservationExclusiveDisable(ReservationExclusive):
    exclusive = forms.BooleanField(
        label=_(u"No overlap"),
        required=False,
        initial=True)

class ConfirmForm(forms.Form):
    confirm = forms.BooleanField()



from distutils.core import setup

setup(name='lyra',
      version='1.1',
      description='time management app for django',
      author='Leo Honkanen',
      author_email='leo.o.honkanen@gmail.com',
      url='https://github.com/hylje/lyra',
      packages=['lyra', 
                'lyra.contrib',
                'lyra.contrib.duty',
                'lyra.contrib.food',
                'lyra.contrib.drive',])



from django.utils.translation import ugettext as _

from lyra import views

class DriveApp(views.Lyra):
    app_name = "drive"

    class crud_app(views.ScheduleApp.crud_app):
        from lyra import forms

        reservation_form = forms.ReservationExclusiveDisable
    
    def get_app_name(self):
        return _(u"Ajonvarauskirja")

    def user_can_view(self, request, reservation=None):
        return request.user.is_authenticated()

    def user_can_edit(self, request, reservation):
        return (super(DriveApp, self).user_can_edit(request, reservation)
                or request.user.has_perm("lyra.change_reservation_lesser"))

    def user_can_delete(self, request, reservation):
        return (super(DriveApp, self).user_can_delete(request, reservation)
                or request.user.has_perm("lyra.delete_reservation_lesser"))

    def user_can_create(self, request):
        return request.user.is_authenticated()
